import Vector
import Numeric
import math
import operator
import ode

class Rope(object):
    def __init__(self, game, player, o2, pos, distance):
        self.player = player
        self.obj = o2
        self.pos = pos
        self.d = distance
        self.objects = [player, o2]

    def segments(self):
        return []

class ODERope(Rope):
    def __init__(self, game, player, o2, pos, distance):
        super(ODERope, self).__init__(game, player, o2, pos, distance)
        geom = game.geometry
        self.geom = geom
        p0 = geom.get_pos(player)
        #print p0, pos
        newbpos = map(lambda x,y: (x+y)/2, p0, pos)
        #print 'spank!'
        v = map(operator.sub, pos,p0)
        #print v
        newb = ode.Body(geom.world)
        m = ode.Mass()
        m.setSphere(1, 10)
        newb.setMass(m)
        newb.setPosition(tuple(pos)+(0,))
        j0 = ode.HingeJoint(geom.world)
        #print o2, o2.static
        if o2.static:
            #print 'attaching to new body'
            body0 = ode.Body(geom.world)
            body0.setPosition(tuple(pos)+(0,))
            j0.attach(body0, newb)
            self.motor = ode.LMotor(geom.world)
            self.body0 = body0
            self.motor.attach(ode.environment, body0)
            self.motor.setNumAxes(2)
            self.motor.setAxis(0, 0, (1, 0, 0))
            self.motor.setAxis(1, 0, (0, 1, 0))
            self.motor.setParam(ode.ParamFMax, 1000000)
            self.motor.setParam(ode.ParamFMax2, 1000000)
            vx, vy = geom.get_velo(o2)
            self.motor.setParam(ode.ParamVel, vx)
            self.motor.setParam(ode.ParamVel2, vy)
            o2p = geom.get_pos(o2)
#            self.motor.setParam(ode.ParamStopERP, 0.8)
            
            self.offset = pos[0]-o2p[0], pos[1]-o2p[1]

        else:
            self.motor = None
            j0.attach(geom.bodies[o2], newb)
        j0.setAxis((0, 0, 1))
        j0.setAnchor(tuple(pos)+(0,))
#        j0.setParam(ode.ParamCFM, 1e-10)
        j1 = ode.SliderJoint(geom.world)
        j1.attach(newb, geom.bodies[player])
        j1.setAxis(tuple(v)+(0,))
        #j1.setAnchor(tuple(newbpos)+(0,))
        j1.setParam(ode.ParamHiStop, 10.0)
        delta = map(operator.sub, pos+[0], geom.bodies[player].getPosition())
        d = math.sqrt(sum(map(operator.mul, delta, delta)))
        j1.setParam(ode.ParamLoStop, -d)
        j1.setParam(ode.ParamStopCFM, 1e-10)
        j1.setParam(ode.ParamStopERP, .9)
        j1.setFeedback(True)

        self.j0 = j0
        self.j1 = j1
        self.newb = newb
        self.oldpos = geom.get_pos(o2)

    def update(self, geom):
        if not self.obj.static: return
        p0x, p0y, p0z = self.body0.getPosition()
        px, py = geom.get_pos(self.obj)
        vx, vy = geom.get_velo(self.obj)
        ox, oy = self.offset
        # body0 is at p0x, p0y, want to be at px+vx+ox, py+vy+oy

        diff = px + vx + ox - p0x, py + vy + oy - p0y
        self.motor.setParam(ode.ParamVel, diff[0])
        self.motor.setParam(ode.ParamVel2, diff[1])
        return

    def segments(self):
        def trim3d(pos):
            return pos[:-1]
        return [(trim3d(self.newb.getPosition()), self.geom.get_pos(self.player))]

    def report(self, game):
        print self.player, game.geometry.bodies[self.player].getPosition(), self.newb.getPosition()
        lo = self.j1.getParam(ode.ParamLoStop)
        hi = self.j1.getParam(ode.ParamHiStop)
        pos = self.newb.getPosition()
        delta = map(operator.sub, pos, game.geometry.bodies[self.player].getPosition())
        d = math.sqrt(sum(map(operator.mul, delta, delta)))
        print self.j1.getParam(ode.ParamStopCFM), self.j1.getParam(ode.ParamStopERP), self.j1.getFeedback(),
        print self.j0.getAnchor(), self.j1.getAxis(), self.j0.getAngle(), self.j0.getAngleRate(), lo, self.j1.getPosition(), hi, self.j1.getPositionRate()
        pass

    def move(self, dist):
        p0x, p0y, p0z = self.body0.getPosition()
        self.body0.setPosition((p0x+dist[0], p0y+dist[1], 0))
        p0x, p0y, p0z = self.newb.getPosition()
        self.newb.setPosition((p0x+dist[0], p0y+dist[1], 0))

class OldRope(object):
    def __init__(self, game, player, o2, pos, distance):
        self.player = player
        self.relpos = Numeric.array([16, 16], pos.typecode())
        self.obj = o2
        self.pos = pos
        self.d = distance

    def check(self, geometry, phys):
        p = geometry.get_pos(self.player) + self.relpos #one endpoint
        v = geometry.get_velo(self.player)
        n = (self.pos - p)
        #taken from: http://local.wasp.uwa.edu.au/~pbourke/geometry/sphereline/
        sum = Numeric.sum
        a = sum(v * v)
        if a == 0:
            return
        b = 2 * sum(v*(self.pos-p))
        c = sum(self.pos*self.pos + p*p - 2*self.pos*p)-self.d*self.d
        if b*b - 4*a*c >= 0:
            u = (-b - math.sqrt(b*b-4*a*c))/(2*a)
            if 0 <= u <= 1:
                # OK, about to cross the circle.
                norm = Vector.normalize(n[Numeric.NewAxis])[0]
                self.pull(geometry, phys, self.player, norm, u)

class BounceRope(OldRope):
    def pull(self, geometry, phys, player, n, u):
        #geometry.set_pos(self.player, self.pos - self.d*n)
        print n, u
        phys.bounce(self.player, 0, self.obj, 0, n, u)

class ExactRope(OldRope):
    def __init__(self, game, player, o2, pos, distance):
        super(ExactRope, self).__init__(player, o2, pos, distance)
        self.start = False
        
    def pull(self, geometry, phys, player, n, u):
        if self.start == False:
            #Cancel velocity perpendicular to normal
            v = geometry.get_velo(player)
            v -= Numeric.dot(v, n)*n
            self.aroundvelo = v.copy()
            if Vector.length(self.aroundvelo) > 0.1:
                self.start = True
        p = geometry.get_pos(player) + self.relpos
        v = geometry.get_velo(player)
        dx, dy = -n # i.e. from position to player
        rad = math.atan2(dy, dx)
        drad = Vector.length(self.aroundvelo)/self.d
        #drad needs to be negative if "clockwise"
        dx, dy = (p+v) - self.pos
        clock = math.atan2(dy, dx)
        if (clock - rad)% (2*math.pi) > math.pi:
            drad = -drad

        #To compensate for flipped y-axis, we reverse degrees.
        newloc = self.pos + self.d*Vector.angle_vector(-math.degrees(rad+drad))
        nv = newloc - p
        geometry.set_velo(player, nv)
